/*
    Copyright 2015-2016 Robert Tari <robert@tari.in>

    SACD Ripper - http://code.google.com/p/sacd-ripper/
    Copyright 2010-2011 by respective authors.

    This file is part of SACD.

    SACD is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    SACD is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SACD.  If not, see <http://www.gnu.org/licenses/gpl-3.0.txt>.
*/

#include "scarletbook.h"

const char* character_set[] =
{
    "US-ASCII",
    "ISO646-JP",
    "ISO-8859-1",
    "SHIFT_JISX0213",
    "KSC5636",
    "GB2312",
    "BIG5",
    "ISO-8859-1"
};

const char* album_genre[] =
{
    "Not used",
    "Not defined",
    "Adult Contemporary",
    "Alternative Rock",
    "Children's Music",
    "Classical",
    "Contemporary Christian",
    "Country",
    "Dance",
    "Easy Listening",
    "Erotic",
    "Folk",
    "Gospel",
    "Hip Hop",
    "Jazz",
    "Latin",
    "Musical",
    "New Age",
    "Opera",
    "Operetta",
    "Pop Music",
    "RAP",
    "Reggae",
    "Rock Music",
    "Rhythm & Blues",
    "Sound Effects",
    "Sound Track",
    "Spoken Word",
    "World Music",
    "Blues"
};

const char* album_category[] =
{
    "Not used",
    "General",
    "Japanese"
};
